
        .title  ARM coprocessor C library support

        // For use with the Yagarto GCC tool chain 4.4.1
        // (C)2010 SPROW

        .global  _swi
        .global  _swix
        .global  _exit
        .global  _sbrk
        .global  _start

        .extern  main


        .section armtubeswis, "ax"
        .code 32

        OS_Exit                 = 0x11
        XOS_Exit                = 0x20011
        OS_CallASWIR12          = 0x71
        XOS_CallASWIR12         = 0x20071
        OS_ChangeEnvironment    = 0x40
        XOS_ChangeEnvironment   = 0x20040

        USRStackSize            = 32768
        V_bit                   = 1 << 28
        C_bit                   = 1 << 29
        Z_bit                   = 1 << 30
        N_bit                   = 1 << 31

_swix:
        // Fall through to _swi() function
        ORR     r0, r0, #0x20000

_swi:
        // Push so stack contains all entry varargs in order
        STMFD   r13!, {r2-r3}

        // Registers used
        STMFD   r13!, {r4-r11, lr}
        MOV     lr, r1
        MOV     r12, r0

        // Select registers as requested by _IN()
        TST     lr, #0x200
        LDRNE   r9, [r13, #36+(9*4)]
        TST     lr, #0x100
        LDRNE   r8, [r13, #36+(8*4)]
        TST     lr, #0x80
        LDRNE   r7, [r13, #36+(7*4)]
        TST     lr, #0x40
        LDRNE   r6, [r13, #36+(6*4)]
        TST     lr, #0x20
        LDRNE   r5, [r13, #36+(5*4)]
        TST     lr, #0x10
        LDRNE   r4, [r13, #36+(4*4)]
        TST     lr, #0x8
        LDRNE   r3, [r13, #36+(3*4)]
        TST     lr, #0x4
        LDRNE   r2, [r13, #36+(2*4)]
        TST     lr, #0x2
        LDRNE   r1, [r13, #36+(1*4)]
        TST     lr, #0x1
        LDRNE   r0, [r13, #36+(0*4)]
        SWI     XOS_CallASWIR12

        // Keep flags and X bit
        MRS     r10, CPSR
        AND     r10, r10, #N_bit | Z_bit | C_bit | V_bit
        AND     r12, r12, #0x20000
        ORR     r12, r10, r12
        
        // Count how many bits were input to find the output address start
        ADD     r10, r13, #36
        MOV     r11, #0x200
b10:
        TST     r11, lr
        ADDNE   r10, r10, #4
        MOVS    r11, r11, LSR#1
        BNE     b10

        // Skip when no output words are expected
        MOVS    r11, lr, LSR#22
        BEQ     b20

        // Save registers as requested by _OUT()
        TST     lr, #0x80000000
        LDRNE   r11, [r10], #4
        STRNE   r0, [r11]
        TST     lr, #0x40000000
        LDRNE   r11, [r10], #4
        STRNE   r1, [r11]
        TST     lr, #0x20000000
        LDRNE   r11, [r10], #4
        STRNE   r2, [r11]
        TST     lr, #0x10000000
        LDRNE   r11, [r10], #4
        STRNE   r3, [r11]
        TST     lr, #0x8000000
        LDRNE   r11, [r10], #4
        STRNE   r4, [r11]
        TST     lr, #0x4000000
        LDRNE   r11, [r10], #4
        STRNE   r5, [r11]
        TST     lr, #0x2000000
        LDRNE   r11, [r10], #4
        STRNE   r6, [r11]
        TST     lr, #0x1000000
        LDRNE   r11, [r10], #4
        STRNE   r7, [r11]
        TST     lr, #0x800000
        LDRNE   r11, [r10], #4
        STRNE   r8, [r11]
        TST     lr, #0x400000
        LDRNE   r11, [r10], #4
        STRNE   r9, [r11]
b20:
        // Check if output flags were requested too
        TST     lr, #0x200000
        LDRNE   r11, [r10]
        STRNE   r12, [r11]

        // Error returns are only 0 or error pointer
        TST     r12, #0x20000
        BEQ     b30
        TST     r12, #V_bit
        MOVEQ   r0, #0
        B       b40
b30:
        // Check the return register number
        MOV     lr, lr, LSR#16
        BIC     lr, lr, #15
        TEQ     lr, #15
        MOVEQ   r0, r12
        STMNEFD r13!, {r0-r9}
        MOVNE   lr, lr, LSL#2
        LDRNE   r0, [r13, lr]
        ADDNE   r13, r13, #4*10
b40:
        // Back to C
        LDMFD   r13!, {r4-r11, lr}
        ADD     r13, r13, #8
        MOV     pc, lr


_exit:
        // Exit
        SWI     OS_Exit

_sbrk:
        // caddr_t _sbrk(int incr)
        // At startup all spare RAM is given anyway so just return heap base
        LDR     r0, =__bss_end__
        MOV     pc, lr

        .section .init, "ax"
_start:
        // Zero init data
        LDR     r1, =__bss_start__
        LDR     r2, =__bss_end__
        MOV     r3, #0
b50:
        TEQ     r1, r2
        STRNE   r3, [r1], #4
        BNE     b50

        // Stack base
        MOV     r0, #14
        MOV     r1, #0
        MOV     r2, #0
        SWI     XOS_ChangeEnvironment

        // Environment RAM limit becomes stack base (rounded down to nearest 8)
        BIC     r13, r1, #7
                
        // Now looks like this
        // +-------------------+ stack base (=application space limit)
        // |      Stack        |
        // +-------------------+ stack limit (=heap limit too)
        // |      Heap         |
        // +-------------------+ heap base (=end of ZI region)
        // |      ZI data      |
        // +-------------------+
        // |      RW data      |
        // +-------------------+
        // | Program & RO data |
        // +-------------------+ &00008000
        
        BL      main
        B       _exit
